#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include <linux/soundcard.h>
#include <linux/cdrom.h>
#include "netaudio.h"

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vprintf (ctl,list);
		va_end (list);
	}
}



enum NETAUDIO_TYPE {
	NET_NOSPECIAL=0,
	NET_AUDIO,
	NET_DSP,
	NET_MIXER,
	NET_FLOPPY,
	NET_CDROM,
};

static NETAUDIO_TYPE tb[256];

#ifdef __GLIBC__
	#define __libc_ioctl __ioctl
#endif

extern "C" {
	int __libc_open (const char *, int, int);
	int __libc_creat (const char *, int);
	int __libc_close(int);
	int __libc_write (int, const void *, unsigned);
	int __libc_read  (int, void *, unsigned);
	int __libc_ioctl (int, int request, ...);
	off_t __libc_lseek(int fd, off_t offset, int whence);
}

static int netaudio_bufread (int cli, void *buf, int len)
{
	char *pbuf = (char*)buf;
	int sofar = 0;
	while (sofar < len){
		int nb=__libc_read(cli,pbuf+sofar,len-sofar);
		if (nb > 0){
			sofar += nb;
		}else if (errno != EAGAIN){
			break;
		}
	}
	return sofar != len ? -1 : len;
}



static int netaudio_readack(int fd, NETAUDIO_CTL &ctl)
{
	int ret = -1;
	if (netaudio_bufread (fd,&ctl,sizeof(NETAUDIO_CTL))!=-1){
		//printf ("tmp = %d %d\n",ctl.len, ctl.p1);
		ret = ctl.len;
		errno = ctl.request;
	}
	return ret;
}
static int netaudio_readack(int fd)
{
	NETAUDIO_CTL ctl;
	return netaudio_readack (fd,ctl);
}

static char audioserver[200];	// Name of the audioserver

static int netaudio_connect (
	const char *fname,
	int mode,
	NETAUDIO_TYPE type)
{
	//printf ("Connect to :%s:\n",audioserver);
	int fd = cmdsock_connect (audioserver,8010,1);
	//printf ("Connect to :%s: -> %d\n",audioserver,fd);
	if (fd != -1){
		NETAUDIO_OPEN op;
		strcpy (op.fname,fname);
		op.mode = mode;
		__libc_write (fd,&op,sizeof(op));
		netaudio_readack(fd);
		tb[fd] = type;
	}
	return fd;
}
enum OPEN_STATE {
	OPEN_UNKNOWN,
	OPEN_LOCAL,
	OPEN_REMOTE
};
static OPEN_STATE state = OPEN_UNKNOWN;

static void netaudio_setstate()
{
	// We check once if we are running in remote ressource mode
	state = OPEN_LOCAL;
	const char *s = getenv ("NETAUDIOHOST");
	if (s == NULL) s = getenv("DISPLAY");
	if (s != NULL){
		strcpy (audioserver,s);
		char *pt = strchr(audioserver,':');
		if (pt != NULL) *pt = '\0';
		if (audioserver != '\0'){
			char hname[PATH_MAX];
			if (gethostname (hname,PATH_MAX-1)!=-1
				&& strcmp(audioserver,hname)!=0){
				state = OPEN_REMOTE;
			}
		}
	}
}

extern "C" int open (const char *fname, int mode, ...)
{
	int ret = -1;
	checkmount (fname);
	if (state == OPEN_UNKNOWN){
		netaudio_setstate();
	}
	if (state == OPEN_REMOTE){
		if (strcmp(fname,"/dev/audio")==0){
			ret = netaudio_connect (fname,mode,NET_AUDIO);
		}else if (strcmp(fname,"/dev/dsp")==0){
			ret = netaudio_connect (fname,mode,NET_DSP);
		}else if (strcmp(fname,"/dev/mixer")==0){
			ret = netaudio_connect (fname,mode,NET_MIXER);
		}else if (strcmp(fname,"/dev/cdrom")==0){
			ret = netaudio_connect (fname,mode,NET_CDROM);
		}else if (strcmp(fname,"/dev/fd0")==0
			|| strcmp(fname,"/dev/fd0H1440")==0
			|| strcmp(fname,"/dev/fd1")==0
			|| strcmp(fname,"/dev/fd1H1440")==0){
			ret = netaudio_connect (fname,mode,NET_FLOPPY);
		}
	}
	if (ret == -1){
		va_list list;
		va_start (list,mode);
		int perm = va_arg(list,int);
		ret = __libc_open (fname,mode,perm);
		va_end (list);
	}
	return ret;
}
#if 0
extern "C" int creat (const char *fname, mode_t mode)
{
	checkmount (fname);
	if (state == OPEN_UNKNOWN){
		netaudio_setstate();
	}
	int ret = -1;
	if (state == OPEN_REMOTE){
		if (strcmp(fname,"/dev/fd0")==0
			|| strcmp(fname,"/dev/fd0H1440")==0
			|| strcmp(fname,"/dev/fd1")==0
			|| strcmp(fname,"/dev/fd1H1440")==0){
			ret = netaudio_connect (fname,mode,NET_FLOPPY);
		}
	}
	if (ret == -1){
		ret = __libc_creat (fname,mode);
	}
	return ret;
}
#endif

extern "C" off_t lseek(int fd, off_t offset, int whence)
{
	NETAUDIO_TYPE type = NET_NOSPECIAL;
	if (fd >=0 && fd < 256) type = tb[fd];
	int ret = -1;
	if (type == NET_NOSPECIAL){
		ret = __libc_lseek (fd,offset,whence);
	}else if (type == NET_DSP
		|| type == NET_MIXER
		|| type == NET_AUDIO){
		errno = EINVAL;
		ret = -1;
	}else if (type == NET_FLOPPY || type == NET_CDROM){
		NETAUDIO_CTL ctl;
		ctl.cmd = NETCTL_LSEEK;
		ctl.p1 = offset;
		ctl.p2 = whence;
		__libc_write (fd,&ctl,sizeof(ctl));
		ret = netaudio_readack(fd);
	}
	return ret;
}
extern "C" int close (int fd)
{
	if (fd >=0 && fd < 256) tb[fd] = NET_NOSPECIAL;
	return __libc_close (fd);
}

extern "C" int write (int fd, const void *buf, unsigned len)
{
	NETAUDIO_TYPE type = NET_NOSPECIAL;
	if (fd >=0 && fd < 256) type = tb[fd];
	int ret;
	if (type == NET_NOSPECIAL){
		ret = __libc_write (fd,buf,len);
	}else{
		NETAUDIO_CTL ctl;
		ctl.cmd = NETCTL_WRITE;
		ctl.len = len;
		__libc_write (fd,&ctl,sizeof(ctl));
		ret = __libc_write (fd,buf,len);
		netaudio_readack(fd);
	}
	return ret;
}

static int netaudio_read (int fd, void *buf, unsigned len)
{
	NETAUDIO_CTL ctl;
	ctl.cmd = NETCTL_READ;
	ctl.len = len;
	int ret = -1;
	if (__libc_write (fd,&ctl,sizeof(ctl))==sizeof(ctl)){
		n_debug ("ecrread ok\n");
		int n = __libc_read (fd,&ctl,sizeof(ctl));
		if (n==(int)sizeof(ctl)){
			int sofar = 0;
			int nb;
			n_debug ("essai de lire %d\n",ctl.len);
			while (sofar < ctl.len
				&& (nb=__libc_read(fd,buf+sofar,len-sofar))>0){
				n_debug ("getting %d , sofar = %d, len = %d\n"
					,nb,sofar,ctl.len);
				sofar += nb;
			}
			n_debug ("obtient %d\n",sofar);
			ret = sofar;
			#if 0
				FILE *fau = fopen ("/tmp/log.au","a");
				fwrite (buf,1,ret,fau);
				fclose (fau);
			#endif
		}else{
			n_debug ("read ctl %d\n",n);
		}
	}else{
		n_debug ("ecrread kaput\n");
	}
	return ret;
}

extern "C" int read (int fd, void *buf, unsigned len)
{
	NETAUDIO_TYPE type = NET_NOSPECIAL;
	if (fd >=0 && fd < 256) type = tb[fd];
	int ret = -1;
	if (type == NET_NOSPECIAL){
		ret = __libc_read (fd,buf,len);
	}else if (type == NET_DSP){
		n_debug ("read from dsp %d\n",len);
		ret = netaudio_read (fd,buf,len);
	}else if (type == NET_AUDIO){
		n_debug ("read from audio %d\n",len);
		ret = netaudio_read (fd,buf,len);
	}else if (type == NET_MIXER){
		n_debug ("read from mixer %d\n",len);
		ret = netaudio_read (fd,buf,len);
	}else if (type == NET_FLOPPY || type == NET_CDROM){
		n_debug ("read from floppy %d\n",len);
		ret = netaudio_read (fd,buf,len);
	}
	return ret;
}

static struct {
	int request;
	const char *name;
	bool seen;
}tbreq[]={
	{SNDCTL_PMGR_IFACE,			"SNDCTL_PMGR_IFACE",	false},
	{SNDCTL_DSP_RESET,			"SNDCTL_DSP_RESET",	false},
	{SNDCTL_DSP_SYNC,			"SNDCTL_DSP_SYNC",	false},
	{SNDCTL_DSP_SPEED,			"SNDCTL_DSP_SPEED",	false},
	{SNDCTL_DSP_STEREO,			"SNDCTL_DSP_STEREO",	false},
	{SNDCTL_DSP_GETBLKSIZE,		"SNDCTL_DSP_GETBLKSIZE",	false},
	{SNDCTL_DSP_SAMPLESIZE,		"SNDCTL_DSP_SAMPLESIZE",	false},
	{SNDCTL_DSP_CHANNELS,		"SNDCTL_DSP_CHANNELS",	false},
	{SOUND_PCM_WRITE_CHANNELS,	"SOUND_PCM_WRITE_CHANNELS",	false},
	{SOUND_PCM_WRITE_FILTER,	"SOUND_PCM_WRITE_FILTER",	false},
	{SNDCTL_DSP_POST,			"SNDCTL_DSP_POST",	false},
	{SNDCTL_DSP_SUBDIVIDE,		"SNDCTL_DSP_SUBDIVIDE",	false},
	{SNDCTL_DSP_SETFRAGMENT,	"SNDCTL_DSP_SETFRAGMENT",	false},
	{SNDCTL_DSP_GETFMTS,		"SNDCTL_DSP_GETFMTS",	false},
	{SNDCTL_DSP_SETFMT,			"SNDCTL_DSP_SETFMT",	false},
	{SNDCTL_DSP_GETOSPACE,		"SNDCTL_DSP_GETOSPACE",	false},
	{SNDCTL_DSP_GETISPACE,		"SNDCTL_DSP_GETISPACE",	false},
	{SNDCTL_DSP_NONBLOCK,		"SNDCTL_DSP_NONBLOCK",	false},
	{SNDCTL_DSP_GETCAPS,		"SNDCTL_DSP_GETCAPS",	false},
	{SNDCTL_DSP_GETTRIGGER,		"SNDCTL_DSP_GETTRIGGER",	false},
	{SNDCTL_DSP_SETTRIGGER,		"SNDCTL_DSP_SETTRIGGER",	false},
	{SNDCTL_DSP_GETIPTR,		"SNDCTL_DSP_GETIPTR",	false},
	{SNDCTL_DSP_GETOPTR,		"SNDCTL_DSP_GETOPTR",	false},
	{SNDCTL_DSP_MAPINBUF,		"SNDCTL_DSP_MAPINBUF",	false},
	{SNDCTL_DSP_MAPOUTBUF,		"SNDCTL_DSP_MAPOUTBUF",	false},
	{SNDCTL_DSP_SETSYNCRO,		"SNDCTL_DSP_SETSYNCRO",	false},
	{SNDCTL_DSP_SETDUPLEX,		"SNDCTL_DSP_SETDUPLEX",	false},
	{SOUND_PCM_READ_RATE,		"SOUND_PCM_READ_RATE",	false},
	{SOUND_PCM_READ_CHANNELS,	"SOUND_PCM_READ_CHANNELS",	false},
	{SOUND_PCM_READ_BITS,		"SOUND_PCM_READ_BITS",	false},
	{SOUND_PCM_READ_FILTER,		"SOUND_PCM_READ_FILTER",	false},

	{SNDCTL_SEQ_RESET,		"SNDCTL_SEQ_RESET",	false},
	{SNDCTL_SEQ_SYNC,		"SNDCTL_SEQ_SYNC",	false},
	{SNDCTL_SYNTH_INFO,		"SNDCTL_SYNTH_INFO",	false},
	{SNDCTL_SEQ_CTRLRATE,	"SNDCTL_SEQ_CTRLRATE",	false},
	{SNDCTL_SEQ_GETOUTCOUNT,	"SNDCTL_SEQ_GETOUTCOUNT",	false},
	{SNDCTL_SEQ_GETINCOUNT,	"SNDCTL_SEQ_GETINCOUNT",	false},
	{SNDCTL_SEQ_PERCMODE,	"SNDCTL_SEQ_PERCMODE",	false},
	{SNDCTL_FM_LOAD_INSTR,	"SNDCTL_FM_LOAD_INSTR",	false},
	{SNDCTL_SEQ_TESTMIDI,	"SNDCTL_SEQ_TESTMIDI",	false},
	{SNDCTL_SEQ_RESETSAMPLES,"SNDCTL_SEQ_RESETSAMPLES",	false},
	{SNDCTL_SEQ_NRSYNTHS,	"SNDCTL_SEQ_NRSYNTHS",	false},
	{SNDCTL_SEQ_NRMIDIS,		"SNDCTL_SEQ_NRMIDIS",	false},
	{SNDCTL_MIDI_INFO,		"SNDCTL_MIDI_INFO",	false},
	{SNDCTL_SEQ_THRESHOLD,	"SNDCTL_SEQ_THRESHOLD",	false},
	{SNDCTL_SEQ_TRESHOLD,	"SNDCTL_SEQ_TRESHOLD",	false},
	{SNDCTL_SYNTH_MEMAVL,	"SNDCTL_SYNTH_MEMAVL",	false},
	{SNDCTL_FM_4OP_ENABLE,	"SNDCTL_FM_4OP_ENABLE",	false},
	{SNDCTL_PMGR_ACCESS,		"SNDCTL_PMGR_ACCESS",	false},
	{SNDCTL_SEQ_PANIC,		"SNDCTL_SEQ_PANIC",	false},
	{SNDCTL_SEQ_OUTOFBAND,	"SNDCTL_SEQ_OUTOFBAND",	false},

	{SOUND_MIXER_READ_VOLUME,	"SOUND_MIXER_READ_VOLUME",false},
	{SOUND_MIXER_READ_BASS,		"SOUND_MIXER_READ_BASS",false},
	{SOUND_MIXER_READ_TREBLE,	"SOUND_MIXER_READ_TREBLE",false},
	{SOUND_MIXER_READ_SYNTH,	"SOUND_MIXER_READ_SYNTH",false},
	{SOUND_MIXER_READ_PCM,		"SOUND_MIXER_READ_PCM",false},
	{SOUND_MIXER_READ_SPEAKER,	"SOUND_MIXER_READ_SPEAKER",false},
	{SOUND_MIXER_READ_LINE,		"SOUND_MIXER_READ_LINE",false},
	{SOUND_MIXER_READ_MIC,		"SOUND_MIXER_READ_MIC",false},
	{SOUND_MIXER_READ_CD,		"SOUND_MIXER_READ_CD",false},
	{SOUND_MIXER_READ_IMIX,		"SOUND_MIXER_READ_IMIX",false},
	{SOUND_MIXER_READ_ALTPCM,	"SOUND_MIXER_READ_ALTPCM",false},
	{SOUND_MIXER_READ_RECLEV,	"SOUND_MIXER_READ_RECLEV",false},
	{SOUND_MIXER_READ_IGAIN,	"SOUND_MIXER_READ_IGAIN",false},
	{SOUND_MIXER_READ_OGAIN,	"SOUND_MIXER_READ_OGAIN",false},
	{SOUND_MIXER_READ_LINE1,	"SOUND_MIXER_READ_LINE1",false},
	{SOUND_MIXER_READ_LINE2,	"SOUND_MIXER_READ_LINE2",false},
	{SOUND_MIXER_READ_LINE3,	"SOUND_MIXER_READ_LINE3",false},

	{SOUND_MIXER_READ_RECSRC,	"SOUND_MIXER_READ_RECSRC",false},
	{SOUND_MIXER_READ_DEVMASK,	"SOUND_MIXER_READ_DEVMASK",false},
	{SOUND_MIXER_READ_RECMASK,	"SOUND_MIXER_READ_RECMASK",false},
	{SOUND_MIXER_READ_STEREODEVS,"SOUND_MIXER_READ_STEREODEVS",false},
	{SOUND_MIXER_READ_CAPS,		"SOUND_MIXER_READ_CAPS",false},


	{SOUND_MIXER_WRITE_VOLUME,	"SOUND_MIXER_WRITE_VOLUME",false},
	{SOUND_MIXER_WRITE_BASS,	"SOUND_MIXER_WRITE_BASS",false},
	{SOUND_MIXER_WRITE_TREBLE,	"SOUND_MIXER_WRITE_TREBLE",false},
	{SOUND_MIXER_WRITE_SYNTH,	"SOUND_MIXER_WRITE_SYNTH",false},
	{SOUND_MIXER_WRITE_PCM,		"SOUND_MIXER_WRITE_PCM",false},
	{SOUND_MIXER_WRITE_SPEAKER,	"SOUND_MIXER_WRITE_SPEAKER",false},
	{SOUND_MIXER_WRITE_LINE,	"SOUND_MIXER_WRITE_LINE",false},
	{SOUND_MIXER_WRITE_MIC,		"SOUND_MIXER_WRITE_MIC",false},
	{SOUND_MIXER_WRITE_CD,		"SOUND_MIXER_WRITE_CD",false},
	{SOUND_MIXER_WRITE_IMIX,	"SOUND_MIXER_WRITE_IMIX",false},
	{SOUND_MIXER_WRITE_ALTPCM,	"SOUND_MIXER_WRITE_ALTPCM",false},
	{SOUND_MIXER_WRITE_RECLEV,	"SOUND_MIXER_WRITE_RECLEV",false},
	{SOUND_MIXER_WRITE_IGAIN,	"SOUND_MIXER_WRITE_IGAIN",false},
	{SOUND_MIXER_WRITE_OGAIN,	"SOUND_MIXER_WRITE_OGAIN",false},
	{SOUND_MIXER_WRITE_LINE1,	"SOUND_MIXER_WRITE_LINE1",false},
	{SOUND_MIXER_WRITE_LINE2,	"SOUND_MIXER_WRITE_LINE2",false},
	{SOUND_MIXER_WRITE_LINE3,	"SOUND_MIXER_WRITE_LINE3",false},
	{SOUND_MIXER_INFO,			"SOUND_MIXER_INFO",false},
	{SOUND_MIXER_ACCESS,		"SOUND_MIXER_ACCESS",false},

	{-1,NULL,	false},
};



static void netaudio_identctl (int request, void *a1, void *a2, void *a3)
{
	int request_ffff = request & 0xffff;
	for (int i=0; tbreq[i].name != NULL; i++){
		//fprintf (stderr,"\t%x<>%x\n",request,tbreq[i].request);
		if ((tbreq[i].request & 0xffff) == request_ffff){
			if (!tbreq[i].seen){
				n_debug ("request %s %p %p %p\n",tbreq[i].name,a1,a2,a3);
				tbreq[i].seen = true;
			}
			break;
		}
	}
}


extern "C" int ioctl (int fd, unsigned request, void *a1, void *a2, void *a3)
{
	NETAUDIO_TYPE type = NET_NOSPECIAL;
	if (fd >=0 && fd < 256) type = tb[fd];
	int ret = -1;
	if (type == NET_NOSPECIAL){
		ret = __libc_ioctl (fd,request,a1,a2,a3);
	}else if (type == NET_DSP){
		netaudio_identctl (request,a1,a2,a3);
		NETAUDIO_CTL ctl;
		ctl.cmd = NETCTL_IOCTL;
		ctl.request = request;
		ctl.p1 = 0;
		ctl.p2 = 0;
		ctl.p3 = 0;
		ret = 0;
		if (request == SNDCTL_DSP_SAMPLESIZE){
			int *p = (int*)a1;
			ctl.p1 = *p;
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
		}else if (request == SNDCTL_DSP_CHANNELS){
			ret = 1;
		}else if (request == SNDCTL_DSP_SPEED){
			int *p = (int*)a1;
			ctl.p1 = *p;
			printf ("speed = %d\n",*p);
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
		}else if (request == SNDCTL_DSP_RESET){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
		}else if (request == SNDCTL_DSP_GETCAPS){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}
	}else if (type == NET_AUDIO){
		n_debug ("ioctl %08x for audio\n",request);
		netaudio_identctl (request,a1,a2,a3);
	}else if (type == NET_MIXER){
		//fprintf (stderr,"ioctl %08x for mixer\n",request);
		NETAUDIO_CTL ctl;
		ctl.cmd = NETCTL_IOCTL;
		ctl.request = request;
		ctl.p1 = 0;
		ctl.p2 = 0;
		ctl.p3 = 0;
		if (request == SOUND_MIXER_READ_CAPS){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_DEVMASK){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_RECMASK){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_STEREODEVS){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_READ_VOLUME
			|| request == SOUND_MIXER_READ_BASS
			|| request == SOUND_MIXER_READ_TREBLE
			|| request == SOUND_MIXER_READ_SYNTH
			|| request == SOUND_MIXER_READ_PCM
			|| request == SOUND_MIXER_READ_SPEAKER
			|| request == SOUND_MIXER_READ_LINE
			|| request == SOUND_MIXER_READ_MIC
			|| request == SOUND_MIXER_READ_CD
			|| request == SOUND_MIXER_READ_IMIX
			|| request == SOUND_MIXER_READ_ALTPCM
			|| request == SOUND_MIXER_READ_RECLEV
			|| request == SOUND_MIXER_READ_IGAIN
			|| request == SOUND_MIXER_READ_OGAIN
			|| request == SOUND_MIXER_READ_LINE1
			|| request == SOUND_MIXER_READ_LINE2
			|| request == SOUND_MIXER_READ_LINE3

			|| request == SOUND_MIXER_READ_RECSRC
			|| request == SOUND_MIXER_READ_DEVMASK
			|| request == SOUND_MIXER_READ_RECMASK
			|| request == SOUND_MIXER_READ_STEREODEVS
			|| request == SOUND_MIXER_READ_CAPS){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else if (request == SOUND_MIXER_WRITE_VOLUME
			|| request == SOUND_MIXER_WRITE_BASS
			|| request == SOUND_MIXER_WRITE_TREBLE
			|| request == SOUND_MIXER_WRITE_SYNTH
			|| request == SOUND_MIXER_WRITE_PCM
			|| request == SOUND_MIXER_WRITE_SPEAKER
			|| request == SOUND_MIXER_WRITE_LINE
			|| request == SOUND_MIXER_WRITE_MIC
			|| request == SOUND_MIXER_WRITE_CD
			|| request == SOUND_MIXER_WRITE_IMIX
			|| request == SOUND_MIXER_WRITE_ALTPCM
			|| request == SOUND_MIXER_WRITE_RECLEV
			|| request == SOUND_MIXER_WRITE_IGAIN
			|| request == SOUND_MIXER_WRITE_OGAIN
			|| request == SOUND_MIXER_WRITE_LINE1
			|| request == SOUND_MIXER_WRITE_LINE2
			|| request == SOUND_MIXER_WRITE_LINE3){
			int *p = (int*)a1;
			ctl.p1 = *p;
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
			*(int*)a1 = ctl.p1;
		}else{
			fprintf (stderr,"unknown ioctl %08x for mixer\n",request);
		}
		netaudio_identctl (request,a1,a2,a3);
	}else if (type == NET_CDROM){
		NETAUDIO_CTL ctl;
		ctl.cmd = NETCTL_IOCTL;
		ctl.request = request;
		ctl.p1 = 0;
		ctl.p2 = 0;
		ctl.p3 = 0;
		//fprintf (stderr,"cdrom request %x\n",request);
		if (request == CDROMSUBCHNL){
			struct cdrom_subchnl *p = (struct cdrom_subchnl*)a1;
			__libc_write (fd,&ctl,sizeof(ctl));
			__libc_write (fd,p,sizeof(*p));
			ret = netaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMREADTOCHDR){
			struct cdrom_tochdr *p = (struct cdrom_tochdr*)a1;
			__libc_write (fd,&ctl,sizeof(ctl));
			__libc_write (fd,p,sizeof(*p));
			ret = netaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMREADTOCENTRY){
			struct cdrom_tocentry *p = (struct cdrom_tocentry*)a1;
			__libc_write (fd,&ctl,sizeof(ctl));
			__libc_write (fd,p,sizeof(*p));
			ret = netaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMPLAYMSF){
			struct cdrom_msf *p = (struct cdrom_msf*)a1;
			__libc_write (fd,&ctl,sizeof(ctl));
			__libc_write (fd,p,sizeof(*p));
			ret = netaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMVOLCTRL
			|| request == CDROMVOLREAD){
			struct cdrom_volctrl *p = (struct cdrom_volctrl*)a1;
			__libc_write (fd,&ctl,sizeof(ctl));
			__libc_write (fd,p,sizeof(*p));
			ret = netaudio_readack(fd,ctl);
			__libc_read  (fd,p,sizeof(*p));
		}else if (request == CDROMPAUSE
			|| request == CDROMRESUME
			|| request == CDROMSTOP
			|| request == CDROMSTART
			|| request == CDROMEJECT
			|| request == CDROMCLOSETRAY
			|| request == CDROMRESET){
			__libc_write (fd,&ctl,sizeof(ctl));
			ret = netaudio_readack(fd,ctl);
		}else if (request == CDROMPLAYTRKIND
			|| request == CDROMREADMODE2
			|| request == CDROMREADMODE1
			|| request == CDROMREADAUDIO
			|| request == CDROMEJECT_SW
			|| request == CDROMMULTISESSION
			|| request == CDROM_GET_UPC
			|| request == CDROMVOLREAD){
			#if 0
				int *p = (int*)a1;
				ctl.p1 = *p;
				__libc_write (fd,&ctl,sizeof(ctl));
				ret = netaudio_readack(fd,ctl);
				*(int*)a1 = ctl.p1;
			#else
				printf ("CDROM ioctl %08x pas fait\n",request);
			#endif
		}else{
			fprintf (stderr,"unknown ioctl %08x for cdrom\n",request);
		}
	}else if (type == NET_FLOPPY){
		fprintf (stderr,"unknown ioctl %08x for floppy\n",request);
		errno = EINVAL;
		ret = -1;
	}
	return ret;
}

#if 0
extern "C" int _fxstat (int ver, int fd, struct stat *buf)
{
	char tmp[100];
	int len = sprintf (tmp,"fstat %d %d\n",ver,fd);
	write (1,tmp,len);
	return -1;
}
#endif
